#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	InstrumentViewer.py
#	
#	Encapsulate the customised track viewing and editing control.
#
#-------------------------------------------------------------------------------

import gtk
import pango
from EventLaneViewer import *
import Globals
import AddInstrumentDialog
import ControlsBox
import Utils
import gettext
_ = gettext.gettext

#=========================================================================	

class InstrumentViewer(gtk.EventBox):
	""" 
	Encapsulates the customized track viewing and editing control.
	"""
	
	""" Size of the instrument name label """
	LABEL_WIDTH_CHARS = 12
	
	""" Widget color when unselected """
	UNSELECTED_COLOUR = None
	
	""" Widget color when selected """
	SELECTED_COLOUR = None
	
	""" Number only to be used inside Jokosher """
	INSTR_DRAG_TYPE = 83
	
	""" Custom numbers for use while dragging audio in Jokosher """
	DRAG_TARGETS = [ ( "jokosher_instr_move", 	# A custom name for the instruments
					   gtk.TARGET_SAME_APP,		# Only move inside Jo
					   INSTR_DRAG_TYPE )]		# Use the custom number
	
	"""
	   The events we wish to receive after we grab the mouse (and it is no longer above this widget)
	   If events other than mouse event are put in here, it may cause the program to crash.
	"""
	_POINTER_GRAB_EVENTS = (
			gtk.gdk.BUTTON_RELEASE_MASK |
			gtk.gdk.BUTTON_PRESS_MASK)
	
	#_____________________________________________________________________
	
	def __init__(self, project, instrument, projectview, mainview, small = False):
		gtk.EventBox.__init__(self)
		"""
		Creates a new instance of InstrumentViewer.
		
		Parameters:
			project -- the currently active Project.
			instrument -- the instrument that the event lane belongs.
			instrumentviewer - the InstrumentViewer holding the event lane.
			projectview - the RecordingView instance that this belongs to.
			mainview - the MainApp Jokosher window.
			small - set to True if we want small edit views (i.e. for mixing view).
		"""
		self.instrument = instrument
		self.project = project
		self.small = small
		self.projectview = projectview
		self.mainview = mainview
		
		self.Updating = False
		
		#get the default colour for the current theme
		self.UNSELECTED_COLOR = self.rc_get_style().bg[gtk.STATE_NORMAL]
		#use base instead of bg colours so that we get the lighter colour that is used for list items in TreeView.
		self.SELECTED_COLOUR = self.rc_get_style().base[gtk.STATE_SELECTED]
		
		self.mainBox = gtk.HBox()
		self.add(self.mainBox)
		
		self.headerBox = gtk.VBox()
		self.headerEventBox = gtk.EventBox()
		self.headerEventBox.add(self.headerBox)
		self.eventLane = EventLaneViewer(project, instrument, self, mainview, self.small)
		
		self.mainBox.pack_start(self.headerEventBox, False, False)
		self.mainBox.pack_end(self.eventLane, True, True)

		# create track header bits
		self.labelbox = gtk.HBox()
		self.labelbox.set_spacing(6)
		self.labeleventbox = gtk.EventBox()
		self.instrlabel = gtk.Label(self.instrument.name)
		self.instrlabel.set_ellipsize(pango.ELLIPSIZE_END)
		self.instrlabel.set_width_chars(self.LABEL_WIDTH_CHARS)
		
		self.editlabel = gtk.Entry()
		self.editlabel.set_text(self.instrument.name)
		self.editlabel.connect("activate", self.OnAcceptEditLabel)
		self.editlabel.connect("key_press_event", self.OnEditLabelKey)
		self.editlabel.connect_after("button-press-event", self.OnEditLabelClick)
		self.editlabel.connect_after("button-release-event", self.OnEditLabelClick)
		self.editlabel_is_mouse_down = False
		self.editlabelPacked = False

		# add the label to the event box
		self.labeleventbox.add(self.instrlabel)

		# set events
		self.labeleventbox.set_events(gtk.gdk.BUTTON_PRESS_MASK | 
						gtk.gdk.ENTER_NOTIFY |
						gtk.gdk.LEAVE_NOTIFY )
		self.labeleventbox.connect("button_press_event", self.OnEditLabel)
		self.labeleventbox.connect("enter_notify_event", self.OnMouseMove)
		self.labeleventbox.connect("leave_notify_event", self.OnMouseMove)
		self.connect("button_press_event", self.OnSelect)
		
		#load missing image icon if instrument has no icon
		if not self.instrument.pixbuf:
			self.instrument.pixbuf = self.render_icon(gtk.STOCK_MISSING_IMAGE, gtk.ICON_SIZE_DIALOG)
		
		self.image = gtk.Image()
		if not (self.small):
			self.image.set_from_pixbuf(self.instrument.pixbuf)
		else:
			pb = self.instrument.pixbuf.scale_simple(20, 20, gtk.gdk.INTERP_BILINEAR)
			self.image.set_from_pixbuf(pb)
		
		self.labelbox.pack_start(self.image, False)
		self.labelbox.pack_end(self.labeleventbox)
		self.headerBox.pack_start(self.labelbox)
		self.controlsBox = ControlsBox.ControlsBox(project,mainview,instrument,includeEffects=True)
		self.headerBox.pack_start(self.controlsBox, False)
		
		self.separator = gtk.HSeparator()
		self.headerBox.pack_end(self.separator, False, True)
		self.instrument.isSelected = False
		
		# Begin Drag and Drop code
		self.headerEventBox.drag_dest_set(gtk.DEST_DEFAULT_MOTION,
										  self.DRAG_TARGETS, 
										  gtk.gdk.ACTION_MOVE)
		self.headerEventBox.connect('drag_motion', self.OnDragMotion)
		self.headerEventBox.drag_source_set(gtk.gdk.BUTTON1_MASK, 
										    self.DRAG_TARGETS, 
										    gtk.gdk.ACTION_MOVE)
		# Connect to drag_begin to add a custom icon
		self.headerEventBox.connect('drag_begin', self.OnDragBegin)
		self.headerEventBox.connect('drag_drop', self.OnDragDrop)
		self.headerEventBox.connect('drag_end', self.OnDragEnd)
		
		self.instrument.connect("name", self.OnInstrumentName)
		self.instrument.connect("image", self.OnInstrumentImage)
		self.instrument.connect("selected", self.OnInstrumentSelected)

		#set the appropriate colour if the instrument it already selected.
		self.OnInstrumentSelected()
		self.show_all()
		self.labelbox.show()
		if self.small:
			self.controlsBox.hide()

	#_____________________________________________________________________
	
	def GetHeaderWidget(self):
		"""
			Returns the widget which is required to be aligned with the instrument headers.
		"""
		return self.headerBox
	
	#_____________________________________________________________________

	def OnSelect(self, widget, event):
		"""
		Called when a button has been pressed anywhere within InstrumentViewer.
		This method sets the instrument to a selected state.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		
		Returns:
			True -- continue GTK signal propagation. *CHECK*
		"""
		
		if self.instrument.project.GetIsRecording():
			return
		
		if 'GDK_CONTROL_MASK' in event.state.value_names:
			self.instrument.SetSelected(True)
		else:
			self.project.ClearEventSelections()
			self.project.SelectInstrument(self.instrument)
		
		return True

	#_____________________________________________________________________

	def OnEditLabel(self, widget, event):
		"""
		Called when a button has been pressed within the instrument name label.
		This method shows a text entry that allows the user to change the name
		of the correspondent Instrument
		
		Parameters:
			widget -- GTK callback.
			event -- GTK callback.
			
		Returns:
			True -- continue GTK signal propagation.
		"""
		if not self.instrument.isSelected:
			# Don't edit label unless the user clicks while we are already selected
			# return false so that the event is handled by the parent widget.
			return False
			
		# replace label with gtk.Entry in order to edit it
		if event.type == gtk.gdk.BUTTON_PRESS:
			#save width of label because when its hidden, width == 0
			width = self.labeleventbox.size_request()[0]
			
			self.editlabel.set_text(self.instrument.name)
			#set the entry to the same width as the label so that the header doesnt resize
			self.editlabel.set_size_request(width, -1)
			self.editlabel.show()
			
			self.labelbox.remove(self.labeleventbox)
			self.labelbox.pack_end(self.editlabel)
			
			self.editlabel.grab_add()
			self.editlabel.grab_focus()
			
			self.editlabelPacked = True
			self.editlabel_is_mouse_down = False
			self.mainview.instrNameEntry = self.editlabel
			
			return True
	
	#_____________________________________________________________________
	
	def OnEditLabelClick(self, widget, event):
		"""
		Handles the button presses while doing a grab on the editing label.
		We need to keep track of both BUTTON_PRESS and BUTTON_RELEASE because:
			1) if the mouse click is on the gtk.Entry, release will be forwarded, press will not.
			2) if the mouse is outside the gtk.Entry, both press and release will be forwarded.
		Therefore if we get both press and release together, we accept the changed and remove the edit dialog.
		"""
		if event.type == gtk.gdk.BUTTON_PRESS:
			self.editlabel_is_mouse_down = True
			return True
		elif event.type == gtk.gdk.BUTTON_RELEASE and self.editlabel_is_mouse_down:
			self.OnAcceptEditLabel()
			return True

	#_____________________________________________________________________

	def OnEditLabelKey(self, widget, event):
		"""
		Handles the key presses while editing the instrument name label.
		Used to make the escape key save the name and then return to normal mode.
		
		Parameters:
			widget -- GTK callback.
			event -- GTK callback.
		"""
		key = gtk.gdk.keyval_name(event.keyval)
		
		if key == "Escape":
			self.editlabel.set_text("")
			self.OnAcceptEditLabel()
			return True

	#_____________________________________________________________________

	def OnAcceptEditLabel(self, widget=None):
		"""
		Called after the instrument label has been edited.
		This method updates the instrument label with the label the user entered.
		
		Parameters:
			widget -- GTK callback.
		"""
		self.editlabel.grab_remove()
		# change instrument name then replace edit label with normal label
		if self.editlabelPacked:
			name = self.editlabel.get_text()
			if name != "":
				self.instrlabel.set_text(name)
			self.labelbox.remove(self.editlabel)
			self.editlabelPacked = False
			self.mainview.instrNameEntry = None
			self.labelbox.pack_end(self.labeleventbox)
			
			if name != "" and name != self.instrument.name:
				#this must be done last because it triggers update
				self.instrument.SetName(name)
		
	#_____________________________________________________________________
	
	def Destroy(self):
		"""
		Called when the InstrumentViewer is closed
		This method also destroys the corresponding EventLaneViewer.
		"""
		self.instrument.disconnect_by_func(self.OnInstrumentImage)
		self.eventLane.Destroy()
		self.destroy()
	
	#_____________________________________________________________________
	
	def OnInstrumentSelected(self, instrument=None):
		"""
		Callback for when the instrument's selected status changes.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		if self.instrument.isSelected:
			self.modify_bg(gtk.STATE_NORMAL, self.SELECTED_COLOUR)
			self.headerEventBox.modify_bg(gtk.STATE_NORMAL, self.SELECTED_COLOUR)
			self.labeleventbox.modify_bg(gtk.STATE_NORMAL, self.SELECTED_COLOUR)
			self.eventLane.modify_bg(gtk.STATE_NORMAL, self.SELECTED_COLOUR)
			
		else:
			self.modify_bg(gtk.STATE_NORMAL, self.UNSELECTED_COLOUR)
			self.headerEventBox.modify_bg(gtk.STATE_NORMAL, self.UNSELECTED_COLOUR)
			self.labeleventbox.modify_bg(gtk.STATE_NORMAL, self.UNSELECTED_COLOUR)
			self.eventLane.modify_bg(gtk.STATE_NORMAL, self.UNSELECTED_COLOUR)
	
	#______________________________________________________________________

	def OnInstrumentName(self, instrument=None):
		"""
		Callback for when the instrument's name changes.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		self.instrlabel.set_text(self.instrument.name)
	
	#_____________________________________________________________________
	
	def OnMouseMove(self, widget, event):
		"""
		Called when the mouse cursor enters or leaves the instrument name label area.
		This method changes the cursor to show text is editable 
		when hovered over the instrument label area.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if not self.window or self.instrument.project.GetIsRecording():
			return
		if (event.type == gtk.gdk.ENTER_NOTIFY):
			self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
		else:
			self.window.set_cursor(None)
			
	#______________________________________________________________________

	def OnDragMotion(self, widget, context, x, y, time):
		"""
		Called each time the user moves the mouse while dragging.
		If the mouse is on an instrument that isn't the source
		instrument, it swaps that instrument and the source instrument
		in the GUI. Swapping of the Instrument objects in self.project.instruments
		happens in OnDragDrop().
		
		Parameters:
			widget -- InstrumentViewer the mouse is hovering over.
			context -- cairo widget context. Used to extract the source instrument.
			x -- reserved for GTK callbacks, don't use it explicitly.
			y -- reserved for GTK callbacks, don't use it explicitly.
			time -- reserved for GTK callbacks, don't use it explicitly.
		
		Returns:
			True -- continue GTK signal propagation. *CHECK*
		"""
		source = context.get_source_widget() 	# Will return an EventBox (self.headerEventBox)
		if widget != source:					# Dont swap with self
			box = self.GetInstrumentViewVBox()
			iv_array = box.get_children()				# InstrumentView array
			index_iv = iv_array.index(self)
			
			source_iv = [iv for iv in iv_array if iv.headerEventBox == source][0]
			index_source_iv = iv_array.index(source_iv)
			
			box.reorder_child(source_iv, index_iv)		# Immediate visual feedback
		# Without these lines the icon would fly back to the start of the drag when dropping
		context.drag_status(gtk.gdk.ACTION_MOVE, time)
		return True

	#______________________________________________________________________
	
	def OnDragBegin(self, widget, context):
		"""
		Called when the drag and drop procedure begins.
		This method will display the instrument icon when dragging.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			context -- a cairo context widget.
		"""
		widget.drag_source_set_icon_pixbuf(self.instrument.pixbuf)
		return True
	
	#______________________________________________________________________
	
	def OnDragDrop(self, widget, context, x, y, time):
		"""
		Called when the user releases MOUSE1, finishing a drag and drop
		procedure. This callback will only be called if the drag is dropped
		on a widget that can handle the drop event. Otherwise only the
		"drag_end" signal will be emitted, This is why the MoveInstrument()
		function is called in OnDragEnd().
			
		Parameters:
			widget -- InstrumentViewer being dragged.
			context -- reserved for GTK callbacks, don't use it explicitly.
			x -- reserved for GTK callbacks, don't use it explicitly.
			y -- reserved for GTK callbacks, don't use it explicitly.
			time -- reserved for GTK callbacks, don't use it explicitly.
		"""
		box = self.GetInstrumentViewVBox()
		position = box.get_children().index(self)
		
		if self.project.instruments.index(self.instrument) == position:
			#there is no change in position
			context.finish(False, False, time)
		else:
			context.finish(True, False, time)
	
	#______________________________________________________________________
	
	def OnDragEnd(self, widget, context):
		"""
		Called when the user releases MOUSE1, finishing a drag and drop
		procedure. This callback will only be called if the drag is dropped
		on a widget that can handle the drop event. Otherwise only the
		"drag_end" signal will be emitted, This is why the MoveInstrument()
		function is called in OnDragEnd().
		Calls MoveInstrument, which moves the dragged instrument to
		the end position in the self.project.instruments array.
		
		Considerations:
			the MoveInstrument action is undo-able.
			
		Parameters:
			widget -- InstrumentViewer being dragged.
			context -- reserved for GTK callbacks, don't use it explicitly.
			x -- reserved for GTK callbacks, don't use it explicitly.
			y -- reserved for GTK callbacks, don't use it explicitly.
			time -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		box = self.GetInstrumentViewVBox()
		position = box.get_children().index(self)
		
		if self.project.instruments.index(self.instrument) != position:
			#there was a change in position
			id = self.instrument.id
			self.project.MoveInstrument(id, position)
		
	#______________________________________________________________________

	def GetInstrumentViewVBox(self):
		"""
		Obtain the current Instrument view box.
		
		Returns:
			the instrumentBox if the current view is a RecordingView.
			the timebox if the current view is a CompactMixView.
		"""
		if hasattr(self.projectview, "instrumentBox"):
			return self.projectview.instrumentBox
		else:
			return self.projectview.timebox
	
	#_____________________________________________________________________
	
	def OnInstrumentImage(self, instrument):
		"""
		Callback for when the instrument's image changes.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		self.image.clear()
		self.image.set_from_pixbuf(self.instrument.pixbuf)
		
	#______________________________________________________________________

	def OnChangeInstrumentType(self, widget, event):
		"""
		Called when a button has been pressed in the instrument header icon.
		This method displays the AddInstrumentDialog, allowing the user
		to change the selected instrument.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
			
		Returns:
			True -- if there's no instrument selected, select one and then
					continue GTK signal propagation. *CHECK*
		"""
		if not self.instrument.isSelected:
			self.OnSelect(widget, event)
			# Don't edit type unless the user clicks while we are already selected
			return True
		
		AddInstrumentDialog.AddInstrumentDialog(self.project, self.mainview, self.instrument)
		    
	#______________________________________________________________________

	def ChangeSize(self, small):
		"""
		Changes the size of the instrument viewer
		
		Parameters:
			small -- True if the instrument viewer is to be small.
		"""
		self.small = small
		self.eventLane.ChangeSize(small)
		if self.small:
			pb = self.instrument.pixbuf.scale_simple(20, 20, gtk.gdk.INTERP_BILINEAR)
			self.image.set_from_pixbuf(pb)
			self.controlsBox.hide()
			self.separator.show()
		else:
			self.image.set_from_pixbuf(self.instrument.pixbuf)
			self.controlsBox.show()
			self.separator.hide()

	#____________________________________________________________________	
#=========================================================================	



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	TimeLine.py
#	
#	This class handles the drawing of the timeline display.
#
#-------------------------------------------------------------------------------

import gtk
import pango
import cairo
import gobject

import gettext
_ = gettext.gettext

#=========================================================================

class TimeLine(gtk.DrawingArea):
	"""
	This class handles drawing the time line display. The time line is part of the
	TimeLineBar. TimeLine displays the time in minutes/seconds (MODE_HOURS_MINS_SECS)
	or bars and beats (MODE_BARS_BEATS). These modes are set in project.transport.
	
	When the time line is constructed in MODE_HOURS_MINS_SECS, it dynamically adjusts
	its scale to the project.viewScale. MODE_BARS_BEATS does not support this (yet).
	"""

	""" GTK widget name """
	__gtype_name__ = 'TimeLine'
	
	""" Number of 'short' lines + 1 (Used for the MODE_HOURS_MINS_SECS timeline)
	Like this:	|				|		   	|
				|  |1 |2 |3 |4 	|  |  |  |  | etc
	"""
	_NUM_LINES = 5
	
	"""
	Various color configurations:
	   ORGBA = Offset, Red, Green, Blue, Alpha
	   RGBA = Red, Green, Blue, Alpha
	   RGB = Red, Green, Blue
	"""
	_BORDER_RGB = (85./255, 85./255, 85./255)
	_BACKGROUND_RGB = (1, 1, 1)
	_TEXT_RGB = (0, 0, 0)
	_BEAT_BAR_RGB = (0, 0, 0)
	_PLAY_CURSOR_RGB = (1, 0, 0)
	
	"""
	   The events we wish to receive after we grab the mouse (and it is no longer above this widget)
	   If events other than mouse event are put in here, it may cause the program to crash.
	"""
	_POINTER_GRAB_EVENTS = (
			gtk.gdk.POINTER_MOTION_MASK |
			gtk.gdk.BUTTON_RELEASE_MASK |
			gtk.gdk.BUTTON_PRESS_MASK)
	
	"""Number of milliseconds between each scrollbar update while autoscrolling"""
	_AUTOSCROLL_UPDATE_INTERVAL = 100
	
	"""
		Speed is relative to distance between mouse and closest edge of the timeline.
		i.e. The further away you pull your mouse, the faster it scrolls.
		The speed, multiplied by the mouse's distance, is the amount we scroll each update.
	"""
	_AUTOSCROLL_SPEED = 0.2
	
	#_____________________________________________________________________

	def __init__(self, project, mainview):
		"""
		Creates a new instance of TimeLine
		
		Parameters:
			project - reference to Project (Project.py)
			mainview - reference to JokosherApp (JokosherApp.py) - Not used atm.
		"""
		gtk.DrawingArea.__init__(self)
	
		self.project = project
		self.mainview = mainview
		
		# Listen for changes in the project and the TransportManager
		self.project.transport.connect("transport-mode", self.OnTransportMode)
		self.project.transport.connect("position", self.OnTransportPosition)
		self.project.connect("bpm", self.OnProjectTimelineChange)
		self.project.connect("time-signature", self.OnProjectTimelineChange)
		self.project.connect("view-start", self.OnProjectTimelineChange)
		self.project.connect("zoom", self.OnProjectTimelineChange)
		
		self.height = 44
		self.buttonDown = False
		self.current_autoscroll_diff = 0

		# source is an offscreen canvas to hold our waveform image
		self.source = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
		
		#rectangle of cached draw area
		self.cachedDrawArea = gtk.gdk.Rectangle(0, 0, 0, 0)

		# Accessibility helpers
		self.SetAccessibleName()
		self.set_property("can-focus", True)
	
		self.set_events(self._POINTER_GRAB_EVENTS)
		self.set_flags(gtk.CAN_FOCUS)
		
		self.connect("expose-event", self.OnDraw)
		self.connect("button_release_event", self.onMouseUp)
		self.connect("button_press_event", self.onMouseDown)
		self.connect("motion_notify_event", self.onMouseMove)
		self.connect("size_allocate", self.OnAllocate)
		
	#_____________________________________________________________________

	def OnAllocate(self, widget, allocation):
		"""
		From:
		http://www.moeraki.com/pygtkreference/pygtk2reference/class-gtkwidget.html#signal-gtkwidget--size-allocate
		The "size-allocate" signal is emitted when widget is given a new space allocation.
		
		Parameters:
			widget -- reserved for GTK callbacks.
			allocation -- the gtk.gdk.Rectangle allocated to the widget.
		"""
		self.allocation = allocation
		
		# Reconstruct the timeline because the size allocation changed
		self.DrawLine()
		
		# Redraw the reconstructed timeline
		self.queue_draw()
		
	#_____________________________________________________________________
		
	def OnDraw(self, widget, event):
		"""
		Fires off the drawing operation.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		area = event.area
		cache = self.cachedDrawArea
		
		#check if the expose area is within the already cached rectangle
		if area.x < cache.x or (area.x + area.width > cache.x + cache.width):
			self.DrawLine()
		
		# Get a cairo surface for this drawing op
		context = widget.window.cairo_create()

		# Give it our timeline image as a source
		context.set_source_surface(self.source, 0, 0)
		# TODO: see what's the problem with using the cached (x,y) values
		#context.set_source_surface(self.source, self.cachedDrawArea.x, self.cachedDrawArea.y)

		# Blit our timeline
		context.paint()
		
		# Draw play cursor position (add 1 so it lines up correctly)
		x = self.project.transport.GetPixelPosition()
		context.set_line_width(1)
		context.move_to(x+0.5, 0)
		context.line_to(x+0.5, self.allocation.height)
		context.set_antialias(cairo.ANTIALIAS_NONE)
		context.set_source_rgb(*self._PLAY_CURSOR_RGB)
		context.stroke()
	
	#_____________________________________________________________________
		
	def DrawLine(self):
		""" 
		Uses Cairo to draw the timeline onto a canvas in memory.
		Must be called initially and to redraw the timeline
		after moving the project start.
		
		Parameters:
			allocation -- the gtk.gdk.Rectangle allocated to the widget.
		"""
		allocArea = self.get_allocation()
		
		#rect = gtk.gdk.Rectangle(allocArea.x - allocArea.width, allocArea.y,
		#				allocArea.width*3, allocArea.height)
		
		# TODO: temporary rect initialization
		rect = allocArea
		
		#Check if our area to cache is outside the allocated area
		#if rect.x < 0:
		#	rect.x = 0
		#if rect.x + rect.width > allocArea.width:
		#	rect.width = allocArea.width - rect.x
		
		self.source = cairo.ImageSurface(cairo.FORMAT_ARGB32, rect.width, rect.height)
		
		context = cairo.Context(self.source)
		context.set_line_width(2)
		context.set_antialias(cairo.ANTIALIAS_NONE)

		# Draw white background
		context.rectangle(0, 0, rect.width, rect.height)
		context.set_source_rgb(*self._BACKGROUND_RGB)
		context.fill()
		
		# Draw the widget border
		context.set_line_width(0.2)
		context.rectangle(0, 0, rect.width, rect.height)
		context.set_source_rgb(*self._BORDER_RGB)
		context.stroke()
		
		x = 0
		transport = self.project.transport
		if transport.mode == transport.MODE_BARS_BEATS:
			# Calculate our scroll offset
			# viewStart is in seconds. Seconds/60 = minutes. Minutes * Beat/Minute = beats (not an integer here)
			pos = (self.project.viewStart / 60.) * self.project.bpm
			
			# floor to an integer. beat = the last beat before viewStart
			beat = int(pos)
			
			# offset = part of a beat that has past since the last beat (offset < 1)
			offset = pos - beat
			
			if offset > 0.:
				# beats * ( pixels/minute ) / ( beats/minute ) = pixels
				# Set x to the position in pixels of the last beat 
				x -= offset * ((self.project.viewScale * 60.) / self.project.bpm)
				
				# (pixels/minute) / (beats/minute) * 1 beat = pixels
				# Add the length of one beat, in pixels
				x += (self.project.viewScale * 60.) / self.project.bpm
				
				# x is now at the pixel-position of the first beat after the viewStart
				beat += 1

			spacing = (60. / self.project.bpm) * self.project.viewScale

			if self.project.meter_denom == 8 and (self.project.meter_nom % 3) == 0 and self.project.meter_nom != 3:
				# Compound time signature, so beats are really 1 dotted note (3 1/8 notes)
				beats_per_bar = self.project.meter_nom / 3
				spacing *= 3
			else:
				# Simple meter
				beats_per_bar = self.project.meter_nom
		
			while x < self.get_allocation().width:
				# Draw the beat/bar divisions
				ix = int(x)

				if beat % beats_per_bar:
					lineHeight = int(self.get_allocation().height/1.2)
				else:
					lineHeight = int(self.get_allocation().height/2)
					
					# Draw the bar number
					context.set_source_rgb(*self._TEXT_RGB)
					number = (beat / beats_per_bar)+1
					
					# TODO: small hack to fix a problem with the numbers not being
					#       properly centered
					if(number == 1):
						context.move_to(ix, 15)
					else:
						context.move_to(ix-3, 15)
						
					context.show_text(str(number))
				
				# Draw the bar itself	
				context.move_to(ix, lineHeight)
				context.line_to(ix, self.get_allocation().height)
				context.set_source_rgb(*self._BEAT_BAR_RGB)
				context.stroke()
					
				beat += 1

				x += spacing
		else:
			# Working in milliseconds here. Using seconds gives modulus problems because they're floats
			viewScale = self.project.viewScale / 1000.
			viewStart = int(self.project.viewStart * 1000)
			factor, displayMilliseconds = self.GetZoomFactor(viewScale)
			
			# Calculate our scroll offset
			# sec : viewStart, truncated to 1000ms; the second that has past just before the beginning of our surface
			msec = viewStart - (viewStart % 1000)
			
			# sec : move to the last 'line' that wasn't drawn
			if (msec % factor) != 0:
				msec -= (msec % factor)
				
			# offset: the amount of milliseconds since the last second before the timeline
			offset = viewStart - msec

			if offset > 0: # x = 0. atm, it should stay that way if offset == 0.
				# offset : milliseconds
				# viewScale : pixels / milliseconds
				# offset * viewScale : offset in pixels
				x -= offset * viewScale # return to the last 'active' second
				x += viewScale * factor # positions the cursor at the first second to be drawn
				msec += factor # cursor is at the first line to be drawn now
				
			# Draw ticks up to the end of our display
			while x < self.get_allocation().width:
				ix = int(x)
				
				if msec % (self._NUM_LINES * factor):
					lineHeight = int(self.get_allocation().height/1.2)
				else:
					lineHeight = int(self.get_allocation().height/2)
					
					# Draw the bar number
					if displayMilliseconds:
						#Should use transportmanager for this...
						number = "%d:%02d:%03d"%((msec/1000) / 60, (msec/1000) % 60, msec%1000)
					else:
						number = "%d:%02d"%((msec/1000) / 60, (msec/1000) % 60)
						
					context.set_source_rgb(*self._TEXT_RGB)
					context.move_to(ix, 15)
					context.show_text(number)
				
				# Draw the bar itself
				context.move_to(ix, lineHeight)
				context.line_to(ix, self.get_allocation().height)
				context.set_source_rgb(*self._BEAT_BAR_RGB)
				context.stroke()
				
				msec += factor
				x += viewScale * factor
		
		#set area to record where the cached surface goes
		self.cachedDrawArea = rect
	
	#_____________________________________________________________________
		
	def do_size_request(self, requisition):
		"""
		From:
		http://www.moeraki.com/pygtkreference/pygtk2reference/class-gtkwidget.html#signal-gtkwidget--size-request
		The "size-request" signal is emitted when a new size is
		requested for widget using the set_size_request() method.
		
		Parameters:
			requisition -- the widget's requested size as a gtk.Requisition.
		"""
		requisition.width = self.get_allocation().width
		requisition.height = self.height
		
	#_____________________________________________________________________
	
	def OnProjectTimelineChange(self, project):
		"""
		Callback for signal when time signature, zoom level,
		bpm or view start of the project change. All of these things
		effect the way that the timeline is drawn.
		
		Parameters:
			project -- the project instance that send the signal.
		"""
		self.queue_draw()
		
	#_____________________________________________________________________
	
	def OnTransportMode(self, transportManager, mode):
		"""
		Callback for signal when the transport mode changes.
		
		Parameters:
			transportManager -- the TransportManager instance that send the signal.
			mode -- the mode type that the transport changed to.
		"""
		self.queue_draw()
	
	#_____________________________________________________________________
	
	def OnTransportPosition(self, transportManager, extraString):
		"""
		Callback for signal when the transport position changes.
		
		Parameters:
			transportManager -- the TransportManager instance that send the signal.
			extraString -- a string specifying the extra action details. i.e. "stop-action"
					means that the position changed because the user hit stop.
		"""
	
		#if the timeline is not currently on screen then quit
		if not self.window:
			return
		
		width_in_secs = self.allocation.width / self.project.viewScale
		# The left and right sides of the viewable area
		rightPos = self.project.viewStart + width_in_secs
		leftPos = self.project.viewStart
		currentPos = self.project.transport.GetPosition()
		
		# Check if the playhead was recently viewable (don't force it in view if it wasn't previously in view)
		# Don't autoscroll if "stop-action" is send in extra because that means the 
		# user just hit stop and did not purposely change the position.
		if "stop-action" != extraString:
			if leftPos < self.project.transport.GetPreviousPosition() < rightPos:
				if currentPos > rightPos:
					# now the playhead has moved off to the right, so force the scroll in that direction
					self.project.SetViewStart(rightPos)
			
				elif currentPos < leftPos:
					#if playhead is beyond leftmost position then force scroll and quit
					leftPos = max(0, leftPos - width_in_secs)
					self.project.SetViewStart(leftPos)
	
		prev_pos = self.project.transport.GetPreviousPixelPosition()
		new_pos = self.project.transport.GetPixelPosition()
	
		self.queue_draw_area(prev_pos - 1, 0, 3, self.get_allocation().height)
		self.queue_draw_area(new_pos - 1, 0, 3, self.get_allocation().height)
	
	#_____________________________________________________________________
	
	def onMouseDown(self, widget, event):
		"""
		Called when a mouse button is clicked.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
			
		Returns:
			True -- to continue the GTK signal propagation.
		"""
		
		response = gtk.gdk.pointer_grab(self.window, False, self._POINTER_GRAB_EVENTS, None, None, event.time)
		self.buttonDown = (response == gtk.gdk.GRAB_SUCCESS)
		self.current_autoscroll_diff = 0
		self.moveHead(event.x)
		self.grab_focus()
		return True

	#_____________________________________________________________________

	def onMouseMove(self, widget, event):
		"""
		Called when the mouse pointer has moved.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly. 
		"""			
		if not self.buttonDown:
			return
		
		old_diff = self.current_autoscroll_diff
		
		alloc = self.get_allocation()
		xpos = event.x
		if 0 < xpos < alloc.width:
			self.current_autoscroll_diff = 0
			self.moveHead(xpos)
		else:
			if xpos > alloc.width:
				self.current_autoscroll_diff = (xpos - alloc.width) * self._AUTOSCROLL_SPEED
			else:
				self.current_autoscroll_diff = xpos * self._AUTOSCROLL_SPEED
			
			if old_diff == 0:
				gobject.timeout_add(self._AUTOSCROLL_UPDATE_INTERVAL, self.onUpdateAutoscroll)
		
	#_____________________________________________________________________
	
	def onUpdateAutoscroll(self):
		if self.current_autoscroll_diff:
			if self.current_autoscroll_diff > 0:
				start_xpos = self.current_autoscroll_diff;
				playhead_xpos = start_xpos + self.allocation.width - 1
			else:
				start_xpos = self.current_autoscroll_diff
				playhead_xpos = start_xpos
				
			start = self.project.viewStart + (start_xpos / self.project.viewScale)
			playhead = self.project.viewStart + (playhead_xpos / self.project.viewScale)
			
			self.project.SetViewStart(start)
			self.project.transport.SeekTo(playhead)

			return True
		else:
			return False
		
	#_____________________________________________________________________
		
	def onMouseUp(self, widget, event):
		"""
		Called when a mouse button is released.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if self.buttonDown:
			self.buttonDown = False
			self.current_autoscroll_diff = 0
			gtk.gdk.pointer_ungrab(event.time)
		
		return True
		
	#_____________________________________________________________________
		
	def moveHead(self, xpos):
		"""
		Changes the project position to the time matching xpos.
		
		Parameters:
			xpos -- the time of the new project position.
		"""
		pos = self.project.viewStart + (xpos / self.project.viewScale)
		pos = max(0., pos)
		self.project.transport.SeekTo(pos)
		self.SetAccessibleName()
		
	#_____________________________________________________________________
	
	def GetZoomFactor(self, viewScale):
		"""
		To be used for drawing the MODE_HOURS_MINS_SECS timeline.
		
		Parameters:
			viewScale -- the view scale in pixels per second.
		
		Returns:
			- an integer factor to be multiplied with the viewScale to zoom the timeline in/out
			- a boolean indicating if milliseconds should be displayed
			The default factor is 1000, meaning that the distance between the short lines of the timeline
			symbolizes 1000 milliseconds. The code will increase of decrease this factor to keep the
			timeline readable. The factors can be set with the zoomLevels array. This array
			contains zoom levels that support precision from 20 ms to 1 minute. More extreme zoom
			levels could be added, but will never be reached because the viewScale is limited.
		"""
		shortTextWidth = 28 # for '0:00' notation
		longTextWidth = 56 # for '0:00:000' notation
		textWidth = shortTextWidth
		whiteSpace = 50
		factor = 1000 # Default factor is 1 second for 1 line
		zoomLevels = [20, 100, 200, 1000, 4000, 12000, 60000]
		if (textWidth + whiteSpace) > (self._NUM_LINES * factor * viewScale):
			factor = zoomLevels[zoomLevels.index(factor) + 1]
			while (textWidth + whiteSpace) > (self._NUM_LINES * factor * viewScale) and factor != zoomLevels[-1]:
				factor = zoomLevels[zoomLevels.index(factor) + 1]
		else:
			while (textWidth + whiteSpace) < (factor * viewScale) and factor != zoomLevels[0]:
				factor = zoomLevels[zoomLevels.index(factor) - 1]
				if factor == 200:
					textWidth = longTextWidth
		return factor, (factor < 200) # 0.2 * 5 = 1.0 second, if the interval is smaller, milliseconds are needed
	
	#_____________________________________________________________________
	
	def SetAccessibleName(self):
		"""
		Set a name property in ATK to help users with screenreaders.
		"""
		if self.project.transport.mode == self.project.transport.MODE_BARS_BEATS:
			tuple_ = self.project.transport.GetPositionAsBarsAndBeats()
			position_text = _("Timeline, %(bars)d bars, %(beats)d beats and %(ticks)d ticks in") % \
					{"bars":tuple_[0], "beats":tuple_[1], "ticks":tuple_[2]}
		else:
			tuple_ = self.project.transport.GetPositionAsHoursMinutesSeconds()
			position_text = _("Timeline, %(hours)d hours, %(mins)d minutes, %(secs)d seconds and %(millis)d milliseconds in") % \
					{"hours":tuple_[0], "mins":tuple_[1], "secs":tuple_[2], "millis":tuple_[3]}

		self.get_accessible().set_name(position_text)

	#_____________________________________________________________________
	
#=========================================================================



#!/usr/bin/env python

from distutils.core import setup
from subprocess import *
import os
import glob

#Create an array with all the locale filenames
I18NFILES = []
for filepath in glob.glob("locale/*/LC_MESSAGES/*.mo"):
	targetpath = os.path.dirname(os.path.join("share/", filepath))
	I18NFILES.append((targetpath, [filepath]))

#Create an array with all the help documents (docbook)
HELPDOCS = []
for filepath in glob.glob("help/jokosher/*/*.xml"):
	targetpath = os.path.dirname(os.path.join("share/gnome/", filepath))
	HELPDOCS.append((targetpath, [filepath]))


#Create an array with all the help images
HELPIMAGES = []
for filepath in glob.glob("help/jokosher/*/figures/*.png"):
	targetpath = os.path.dirname(os.path.join("share/gnome/", filepath))
	HELPIMAGES.append((targetpath, [filepath]))
	
#Check if scrollkeeper is available
OMFFILES = []
omfdir = None
try:
	process = Popen(args=["scrollkeeper-config", "--omfdir"], stdout=PIPE)
except OSError:
	#Not available. Skip the registration.
	pass
else:
	#Obtain the OMF repository directory to install and register the help files
	omfdir = os.path.join(process.stdout.read().strip(), "jokosher")
	OMFFILES.append((omfdir, glob.glob("help/jokosher/*.omf")))
		
dist = setup(name='jokosher',
	version='0.11.3',
	author='Jokosher Project',
	author_email='jokosher-devel-list@gnome.org',
	maintainer='Jokosher Project',
	maintainer_email='jokosher-devel-list@gnome.org',
	description='Multi-track non-linear audio editing.',
	long_description='Jokosher is a simple yet powerful multi-track studio. With it you can create and record music, podcasts and more, all from an integrated simple environment.',
	url='http://www.jokosher.org/',
	download_url='http://www.jokosher.org/download',
	license='GNU GPL',
	platforms='linux',
	scripts=['bin/jokosher'],
	packages=['Jokosher', 'Jokosher/elements', 'Jokosher/ui', 'Jokosher/PlatformUtils'],
	data_files=[
		('share/jokosher/', glob.glob("Jokosher/*.glade")),
		('share/jokosher/', ["Jokosher/jokosher-logo.png"]),
		('share/jokosher/Instruments', glob.glob("Instruments/*.instr")),
		('share/jokosher/Instruments/images', glob.glob('Instruments/images/*.png')),
		('share/icons/hicolor/48x48/apps', ['images/jokosher.png']),
		('share/pixmaps', ['images/jokosher.png']),
		('share/jokosher/pixmaps', glob.glob("images/*.png")),
		('share/applications', ['bin/jokosher.desktop']),
		('share/mime/packages',	['bin/jokosher.xml']),
		('share/jokosher/extensions', glob.glob("extensions/*.py") + glob.glob("extensions/*.egg"))
		]+I18NFILES+HELPDOCS+HELPIMAGES+OMFFILES
)

#Non-documented way of getting the final directory prefix
installCmd = dist.get_command_obj(command="install_data")
installdir = installCmd.install_dir
installroot = installCmd.root

if not installroot:
	installroot = ""

if installdir:
	installdir = os.path.join(os.path.sep,
			installdir.replace(installroot, ""))

# Update the real URL attribute inside the OMF files
# and register them with scrollkeeper
if omfdir != None and installdir != None and dist != None:
	
	#Create an array with the docbook file locations
	HELPURI = []
	for filepath in glob.glob("help/jokosher/*/jokosher.xml"):
		targeturi = os.path.join(installdir, "share/gnome/", filepath)
		HELPURI.append(targeturi)
	
	#Replace the URL placeholder inside the OMF files
	installedOmfFiles = glob.glob(installroot + omfdir + "/*.omf")
	for fileNum in range(0, len(installedOmfFiles)):
		call(["scrollkeeper-preinstall", HELPURI[fileNum],
			installedOmfFiles[fileNum], installedOmfFiles[fileNum]])
		
	#Update the scrollkeeper catalog
	if os.geteuid() == 0:
		print "Updating the scrollkeeper index..."
		call(["scrollkeeper-update", "-o", installroot + omfdir])

# Update the mime types
if os.geteuid() == 0 and dist != None:
	print "Updating the mime-types...."
	
	#update the mimetypes database
	try:
	    call(["update-mime-database", "/usr/share/mime/"])
	except:
	    pass
	
	#update the .desktop file database
	try:
	   call(["update-desktop-database"])
	except:
	    pass

print "\nInstallation finished! You can now run Jokosher by typing 'jokosher' or through your applications menu icon."
	
## To uninstall manually delete these files/folders:
## /usr/bin/jokosher
## /usr/share/jokosher/
## /usr/share/gnome/help/jokosher/
## /usr/share/icons/hicolor/48x48/apps/jokosher.png
## /usr/share/locale/*/LC_MESSAGES/jokosher.mo
## /usr/share/pixmaps/jokosher.png
## /usr/share/applications/jokosher.desktop
## /usr/lib/python2.X/site-packages/Jokosher/
## /usr/share/omf/jokosher/*.omf



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	This module handles the creation of mixdown actions.
#	
#
#-------------------------------------------------------------------------------

import os
import gtk, gtk.glade
import gobject

import Globals
import ProjectManager

import gettext
_ = gettext.gettext


class MixdownActionException(Exception):
	"""
	An exception that occurs when there is a problem when the RunAction() method
	in a MixdownAction. The 'message' attribute is a string to display to the user.
	"""
	def __init__(self, message):
		Exception.__init__(self, message)

#=========================================================================

class RegisterMixdownActionAPI:
	"""
	This class handles the registering and deregistering of MixdownActions.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self):
		"""
		Creates a new instance of RegisterMixdownActionAPI.
		"""
		self.registeredActions = []
		
	#_____________________________________________________________________
	
	def RegisterMixdownActions(self, mixdownActions):
		"""
		Called when MixdownActions needs to be registered.
		Appends the MixdownActions specified to the registered mixdown action list (self.registeredActions)
	
		Parameters:
			mixdownActions -- reference to a tuple containing MixdownAction objects
		"""
		for action in mixdownActions:
			self.registeredActions.append(action)
		
	#_____________________________________________________________________

	def DeregisterMixdownActions(self, mixdownActions):
		"""
		Called when MixdownActions needs to be deregistered.
		Removes the MixdownActions specified from the registered mixdown action list (self.registeredActions)
	
		Parameters:
			mixdownAction -- reference to a tuple containing MixdownAction objects
		"""
		for action in mixdownActions:
			self.registeredActions.remove(action)
	
	#_____________________________________________________________________
	
	def ReturnAllActions(self):
		"""
		Returns all MixdownActions in the registered mixdown action list (self.registeredActions)
		"""
		return self.registeredActions
	
	#_____________________________________________________________________

#=========================================================================

class MixdownAction(gobject.GObject):
	"""
	Represents a mixdown action, used as a building block for mixdown profiles.
	"""

	"""
	Signals:
		"action-configured" -- MixdownAction has been configured
	"""
	
	__gsignals__ = {
		"action-configured" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () )
	}

	#_____________________________________________________________________
	
	def __init__(self):
		"""
		Creates a new instance of MixdownAction.
		"""
		gobject.GObject.__init__(self)
		
		# A MixdownAction object has a name, description and icon path.
		self.name = None
		self.description = None
		self.iconPath = None
		
		self.config = {}
		self.isConfigured = None
		
		# to be used by the configuration window in MixdownAction sublcasses.
		self.dialogIcon = gtk.gdk.pixbuf_new_from_file( os.path.join(Globals.IMAGE_PATH, "jokosher.png") )

	
	#_____________________________________________________________________
	
	def ConfigureAction(self):
		"""
		Subclasses should override this method to do configuration.
		This method should handle the configuration of MixdownActions, such as
		modifying details associated with the MixdownAction.
		"""
		pass
		
	#_____________________________________________________________________
	
	def RunAction(self):
		"""
		Do whatever this action actually does.
		"""
		pass
		
	#_____________________________________________________________________
	
#=========================================================================

class ExportAsFileType(MixdownAction):
	"""
	Defines a mixdown action that exports the audio in the user's project to a given file.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self, project):
		"""
		Special init function for this particular MixdownAction, which takes
		a reference to the project as a parameter; this is so we can call back
		into the project to do the export.
		This means that the core code must special-case this action creation,
		and if it's creating one of these then pass it a project.
		
		Parameters:
			project -- the current Jokosher Project.
		"""
		MixdownAction.__init__(self)
		self.name = _("Export File")
		self.description = _("Export a file to the destination of your choosing.")
		# stock gtk icon
		self.iconPath = gtk.STOCK_JUMP_TO
		
		self.project = project # note: this is special to this action, and needs
		                       # to be passed, so this action needs special handling
		                       # in the core.
	
	#_____________________________________________________________________
	
	def ConfigureAction(self):
		"""
		See MixdownAction.ConfigureAction
		"""
		self.configureExportTree = gtk.glade.XML(Globals.GLADE_PATH, "ConfigureExportFileAction")
		signals = {
			"on_cancel_button_clicked" : self.__OnCancel,
			"on_ok_button_clicked" : self.__OnOK,
			"on_local_radio_toggled" : self.__OnLocalRadioToggled,
			"on_server_radio_toggled" : self.__OnServerRadioToggled
		}
			
		self.configureExportTree.signal_autoconnect(signals)
		
		self.configureExportWindow = self.configureExportTree.get_widget("ConfigureExportFileAction")
		self.fileEntry = self.configureExportTree.get_widget("file_entry")
		self.formatCombo = self.configureExportTree.get_widget("format_combo")
		self.localRadio = self.configureExportTree.get_widget("local_radio")
		self.serverRadio = self.configureExportTree.get_widget("server_radio")
		self.detailsVBox = self.configureExportTree.get_widget("details_vbox")
		
		self.formatModel = gtk.ListStore(str, str, str) # description, extension, pipeline
			
		# set some properties
		self.configureExportWindow.set_icon(self.dialogIcon)
		self.formatCombo.set_model(self.formatModel)
		self.formatCombo.clear()
		desc = gtk.CellRendererText()
		self.formatCombo.pack_start(desc)
		self.formatCombo.add_attribute(desc, "text", 0)

		# construct the configuration widgets
		self.__ConstructTable()
		self.__CreateWidgets()
		self.__AttachWidgets()
		self.__BuildLocalHBox()
		self.__PopulateFormatComboBox()
		self.__LoadConfiguration()

		if self.localRadio.get_active():
			self.detailsVBox.pack_start(self.localHBox, False, False)
		else:
			self.detailsVBox.pack_start(self.serverTable, False, False)
		self.detailsVBox.show_all()
		
	#_____________________________________________________________________

	def RunAction(self):
		"""
		See MixdownAction.RunAction
		"""
		filePath = os.path.join( self.config["location"], "%s.%s" % ( self.config["filename"], self.config["filetype"] ) )
		try:
			self.project.Export(filePath, self.config["pipeline"])
		except ProjectManager.ProjectExportException, e:
			msg = "%s"
			if e.errno == ProjectManager.ProjectExportException.MISSING_ELEMENT:
				msg = _("An encoding plugin could not be found: %s.\nPlease make sure all the plugins required for the specified format are installed.")
			elif e.errno == ProjectManager.ProjectExportException.INVALID_ENCODE_BIN:
				msg = _("The encoding bin description is invalid: %s")
			msg %= e.message
			raise MixdownActionException(msg)
		
	#_____________________________________________________________________
	
	def __ConstructTable(self):
		"""
		Called when the server table (self.serverTable) needs to be constructed.
		Creates the server table.
		"""
		self.serverTable = gtk.Table(4, 2)
		self.serverTable.set_row_spacing(0, 6)
		self.serverTable.set_row_spacing(1, 6)
		self.serverTable.set_row_spacing(2, 6)
		self.serverTable.set_row_spacing(3, 6)
		self.serverTable.set_col_spacing(0, 12)
		self.serverTable.set_col_spacing(1, 12)
		
	#_____________________________________________________________________
	
	def __CreateWidgets(self):
		"""
		Called when widgets in the server table (self.serverTable) need to be created.
		Creates the widgets associated with the server table.
		"""
		self.serverType = gtk.Label(_("Server type:"))
		self.serverType.set_alignment(0, 0.5)
		
		self.serverTypeCombo = gtk.ComboBox()
		self.serverTypeCombo.clear()
		textrend = gtk.CellRendererText()
		self.serverTypeCombo.pack_start(textrend)
		self.serverTypeCombo.add_attribute(textrend, "text", 0)
		self.serverTypeModel = gtk.ListStore(str)
		self.serverTypeModel.append(("HTTP",))
		self.serverTypeModel.append(("FTP",))
		self.serverTypeCombo.set_model(self.serverTypeModel)
		
		self.serverLocation = gtk.Label(_("Server location:"))
		self.serverLocation.set_alignment(0, 0.5)
		self.serverLocationEntry = gtk.Entry()
		
		self.username = gtk.Label(_("Username:"))
		self.username.set_alignment(0, 0.5)
		self.usernameEntry = gtk.Entry()
		
		self.password = gtk.Label(_("Password:"))
		self.password.set_alignment(0, 0.5)
		self.passwordEntry = gtk.Entry()
		self.passwordEntry.set_visibility(False)
		
	#_____________________________________________________________________

	def __AttachWidgets(self):
		"""
		Called when the server widgets need to be added to the server table (self.serverTable).
		Attaches the server widgets to the server table.
		"""
		self.serverTable.attach(self.serverType, 0, 1, 0, 1)
		self.serverTable.attach(self.serverTypeCombo, 1, 2, 0, 1)
		self.serverTable.attach(self.serverLocation, 0, 1, 1, 2)
		self.serverTable.attach(self.serverLocationEntry, 1, 2, 1, 2)
		self.serverTable.attach(self.username, 0, 1, 2, 3)
		self.serverTable.attach(self.usernameEntry, 1, 2, 2, 3)
		self.serverTable.attach(self.password, 0, 1, 3, 4)
		self.serverTable.attach(self.passwordEntry, 1, 2, 3, 4)
		
	#_____________________________________________________________________

	def __BuildLocalHBox(self):
		"""
		Called when the location hbox needs to be created.
		Sets up the location hbox in configuration dialog.
		"""
		self.localHBox = gtk.HBox()
		self.localHBox.set_spacing(6)
		self.localEntry = gtk.Entry()
		self.localButton = gtk.Button()
		self.localButton.connect("clicked", self.__ShowDirectoryChooserDialog)
		self.openImage = gtk.Image()
		self.openImage.set_from_stock(gtk.STOCK_OPEN, gtk.ICON_SIZE_BUTTON)
		self.localButton.set_image(self.openImage)
		self.localHBox.pack_start(gtk.Label(_("Location:")), False, False)
		self.localHBox.pack_start(self.localEntry, True, True)
		self.localHBox.pack_start(self.localButton, False, False)
	
	#_____________________________________________________________________
	
	def __PopulateFormatComboBox(self):
		"""
		Called when the format combo model (self.formatModel) needs to be populated.
		Populates the format combo model with Jokosher's currently available export formats.
		"""
		for format in Globals.EXPORT_FORMATS:
			self.formatModel.append((format["description"], format["extension"], format["pipeline"]))
				
	#_____________________________________________________________________
	
	def __LoadConfiguration(self):
		"""
		Called when the configuration for the configuration window needs to be set.
		Updates the widgets to reflect the configuration details that have been loaded.
		"""
		if self.config.has_key("filename"):
			self.fileEntry.set_text(self.config["filename"])
		if self.config.has_key("location"):
			self.localEntry.set_text(self.config["location"])
		if self.config.has_key("filetype"):
			active = -1
			for item in self.formatModel:
				active += 1
				if item[1] == self.config["filetype"]:
					self.formatCombo.set_active(active)
					break

	#_____________________________________________________________________
	
	def __FinishConfiguration(self):
		"""
		Called when configuration needs to be finished.
		This method finishes the action's configuration by modifying
		the configuration dictionary (self.config) to use the new
		configuration details.
		"""
		self.config["filename"] = self.fileEntry.get_text()
		self.config["location"] = self.localEntry.get_text()
		self.config["filetype"] = self.formatModel[self.formatCombo.get_active()][1]
		self.config["pipeline"] = self.formatModel[self.formatCombo.get_active()][2]
		
		fullPath = os.path.join( self.config["location"], "%s.%s" % ( self.config["filename"], self.config["filetype"] ) )
		
		if fullPath in self.project.GetInputFilenames():
			directory, fileName, ext = self.__MakeUniqueFilename(fullPath)
			self.config["filename"] = fileName
			self.config["location"] = directory
			self.config["filetype"] = ext.split(".")[1]
			
		Globals.settings.general["projectfolder"] = os.path.dirname(fullPath)
		Globals.settings.write()

	#_____________________________________________________________________
	
	def __ShowDirectoryChooserDialog(self, widget):
		"""
		Shows a gtk.FileChooserDialog, allowing the user
		to select which directory their audio should be exported to.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		chooser = gtk.FileChooserDialog((_('Select Export Location')), None, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))
		if os.path.exists(Globals.settings.general["projectfolder"]):
			chooser.set_current_folder(Globals.settings.general["projectfolder"])
		else:
			chooser.set_current_folder(os.path.expanduser("~"))
		chooser.set_icon(self.dialogIcon)

		chooser.set_default_response(gtk.RESPONSE_OK)
		chooser.set_transient_for(self.configureExportWindow)
		response = chooser.run()
			
		if response == gtk.RESPONSE_OK:
			filename = chooser.get_filename()
			self.localEntry.set_text(filename)
		chooser.destroy()
		
	#_____________________________________________________________________
	
	def __OnCancel(self, widget):
		"""
		Called when the Cancel Button is clicked.
		Destroys the configuration window.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.configureExportWindow.destroy()
		
	#_____________________________________________________________________
		
	def __OnOK(self, widget):
		"""
		Called when the OK Button is clicked.
		Finishes the mixdown action's configuration.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		# the local directory is the only option which works right now
		if self.localRadio.get_active():
			if self.fileEntry.get_text() and self.localEntry.get_text():
				self.__FinishConfiguration()
		
		self.emit("action-configured")
		self.configureExportWindow.destroy()
	
	#_____________________________________________________________________
	
	def __OnLocalRadioToggled(self, widget):
		"""
		Called when the local directory radio button is activated.
		Shows the location hbox.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		for children in self.detailsVBox.get_children():
			self.detailsVBox.remove(children)
		self.detailsVBox.pack_start(self.localHBox, False, False)
		self.detailsVBox.show_all()

	#_____________________________________________________________________

	def __OnServerRadioToggled(self, widget):
		"""
		Called when the server radio button is activated.
		Shows the server table.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		for children in self.detailsVBox.get_children():
			self.detailsVBox.remove(children)
		self.detailsVBox.pack_start(self.serverTable, False, False)
		self.detailsVBox.show_all()
		
	#_____________________________________________________________________
	
	def __MakeUniqueFilename(self, filename):
		"""
		From a given filename generates a name which doesn't exist
		by appending increasing numbers to it as necessary.
		
		Return:
			the unique filename
		"""
		dirName, baseName = os.path.split(filename)
		name, ext =  os.path.splitext(baseName)
		current = name
		count = 1
		while os.path.exists(os.path.join(dirName, current + ext)):
			current = "%s-%d" % (name, count)
			count += 1
		return (dirName, current, ext)
		
	#_____________________________________________________________________

#=========================================================================

class RunAScript(MixdownAction):
	"""
	Defines a mixdown action that runs a given script while mixing down.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self):
		"""
		See MixdownAction.__init__
		"""
		MixdownAction.__init__(self)
		self.name = _("Run External Script")
		self.description = _("Run an external script from the destination of your choosing.")
		self.iconPath = os.path.join(Globals.IMAGE_PATH, "effect_miscellaneous.png")

	#_____________________________________________________________________

	def ConfigureAction(self):
		"""
		See MixdownAction.ConfigureAction.
		"""
		chooser = gtk.FileChooserDialog((_("Run External Script")), None, gtk.FILE_CHOOSER_ACTION_OPEN, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))
		if os.path.exists(Globals.settings.general["projectfolder"]):
			chooser.set_current_folder(Globals.settings.general["projectfolder"])
		else:
			chooser.set_current_folder(os.path.expanduser("~"))
		chooser.set_icon(self.dialogIcon)

		chooser.set_default_response(gtk.RESPONSE_OK)
		response = chooser.run()
			
		if response == gtk.RESPONSE_OK:
			filename = chooser.get_filename()
			self.config["script"] = filename
			self.emit("action-configured")
		chooser.destroy()
	
	#_____________________________________________________________________

	def RunAction(self):
		"""
		See MixdownAction.RunAction.
		"""
		import subprocess
		import os.path
		# Interestingly when you use shell=True Popen doesn't raise an OSError if the script doesn't exist
		# because the shell runs happily (you get a shell error on stderr instead).
		# So we need to raise it by hand, or use shell=False and require an interpreter line
		if os.path.exists(self.config["script"]):
			try:
				# needs to happen in a thread!
				subprocess.Popen(self.config["script"] ,shell=True).wait()
			except OSError:
				Globals.debug(_("An error occured with the script %s") % self.config["script"])
			except:
				Globals.debug(_("Error in script %s") % self.config["script"])
		
		else:
			Globals.debug(_("The script %s does not exist.") % self.config["script"])
				
	#_____________________________________________________________________
	
#=========================================================================




#
#       THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#       THE 'COPYING' FILE FOR DETAILS
#
#       ControlsBox.py
#       
#	Contains instrument buttons common to the InstrumentViewer and MixerStrip classes.
#
#-------------------------------------------------------------------------------

import pygtk
pygtk.require("2.0")
import gtk
import os.path
import Globals
import Utils
import InstrumentEffectsDialog
import gettext
_ = gettext.gettext

#=========================================================================

class ControlsBox(gtk.HBox):

	def __init__(self, project, mainview, instrument, includeEffects=True):
		gtk.Container.__init__(self)
		"""
		Creates a new instance of ControlsBox.

		Parameters:
			mainview -- The Main Jokosher window 
			instrument -- The instrument that the ControlsBox control.
			includeEffects -- Whether the Effects button is included or not.
			This may be extended to include all the buttons.
		"""

		self.project = project
		self.mainview = mainview
		self.instrument = instrument

		# define the tooltip messages and images for buttons that change states
		self.recTipDisabled = _("Enable this instrument for recording")
		self.recTipEnabled = _("Disable this instrument for recording")
		self.muteTipDisabled = _("Mute - silence this instrument")
		self.muteTipEnabled = _("Unmute - hear this instrument")
		self.soloTipDisabled = _("Activate Solo - silence all other instruments")
		self.soloTipEnabled = _("Deactivate Solo - hear all the instruments")
		
		self.recImgDisabled = gtk.gdk.pixbuf_new_from_file(os.path.join(Globals.IMAGE_PATH, "icon_arm.png"))
		self.recImgEnabled = gtk.gdk.pixbuf_new_from_file(os.path.join(Globals.IMAGE_PATH, "icon_disarm.png"))
		self.soloImgDisabled = gtk.gdk.pixbuf_new_from_file(os.path.join(Globals.IMAGE_PATH, "icon_solo.png"))
		self.soloImgEnabled = gtk.gdk.pixbuf_new_from_file(os.path.join(Globals.IMAGE_PATH, "icon_group.png"))
		self.muteImgDisabled = Utils.GetIconThatMayBeMissing("stock_volume", gtk.ICON_SIZE_BUTTON, False)
		self.muteImgEnabled = Utils.GetIconThatMayBeMissing("stock_volume-mute", gtk.ICON_SIZE_BUTTON, False)
		
		self.recTip = gtk.Tooltips()
		self.recButton = gtk.ToggleButton()
		self.recTip.set_tip(self.recButton, self.recTipEnabled, None)
		self.recButton.connect("toggled", self.OnArm)
		
		self.muteButton = gtk.ToggleButton()
		self.muteButton.connect("toggled", self.OnMute)
		self.muteTip = gtk.Tooltips()
		self.muteTip.set_tip(self.muteButton, self.muteTipDisabled, None)
		
		self.soloButton = gtk.ToggleButton()
		self.soloTip = gtk.Tooltips()
		self.soloTip.set_tip(self.soloButton, self.soloTipDisabled, None)
		self.soloButton.connect("toggled", self.OnSolo)
		
		self.add(self.recButton)
		self.add(self.muteButton)
		self.add(self.soloButton)

		if includeEffects:
			self.propsButton = gtk.Button()
			procimg = gtk.Image()
			procimg.set_from_file(os.path.join(Globals.IMAGE_PATH, "icon_effectsapply.png"))
			self.propsButton.set_image(procimg)
			self.effectsDialog = None		#the instrument effects dialog (to make sure more than one is never opened)

			self.propsButton.connect("clicked", self.OnEffectsButtonClicked)
			self.propsTip = gtk.Tooltips()
			self.propsTip.set_tip(self.propsButton, _("Instrument Effects"), None)
			self.add(self.propsButton)
		
		self.instrument.connect("solo", self.OnInstrumentSolo)
		self.instrument.connect("arm", self.OnInstrumentArm)
		self.instrument.connect("mute", self.OnInstrumentMute)
		
		#initialize the images on the buttons
		for i in (self.OnInstrumentArm, self.OnInstrumentMute, self.OnInstrumentSolo):
			i(self.instrument)

	#_____________________________________________________________________

	def OnMute(self, widget):
		"""
		Toggles muting the instrument on/off.
		It will also update the pressed in/out look of the button.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if not self.Updating:
			self.instrument.ToggleMuted(wasSolo=False)
	
	#_____________________________________________________________________

	def OnArm(self, widget):
		"""
		Toggles arming the instrument on/off.
		It will also update the pressed in/out look of the button.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if not self.Updating:
			self.instrument.ToggleArmed()
		
	#_____________________________________________________________________
	
	def OnSolo(self, widget):
		"""
		Toggles soloing the instrument on/off.
		It will also update the pressed in/out look of the button.
		"""
		if not self.Updating:
			self.instrument.ToggleSolo(False)
	#_____________________________________________________________________

	def OnInstrumentSolo(self, instrument=None):
		"""
		Callback for when the instrument's solo status changes.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		self.Updating = True
		self.soloButton.set_active(self.instrument.isSolo)
		self.Updating = False
		self.soloTip.enable()
		
		# update the solo button image and tooltip
		image = gtk.Image()
		if self.instrument.isSolo:
			image.set_from_pixbuf(self.soloImgEnabled)
			self.soloButton.set_image(image)
			self.soloTip.set_tip(self.soloButton, self.soloTipEnabled, None)
		else:
			image.set_from_pixbuf(self.soloImgDisabled)
			self.soloButton.set_image(image)
			self.soloTip.set_tip(self.soloButton, self.soloTipDisabled, None)

	#_____________________________________________________________________
	
	def OnInstrumentArm(self, instrument=None):
		"""
		Callback for when the instrument's armed status changes.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		self.Updating = True
		self.recButton.set_active(self.instrument.isArmed)
		self.Updating = False
		self.recTip.enable()
		
		# update the arm button image and tooltip	
		image = gtk.Image()
		if self.instrument.isArmed:
			image.set_from_pixbuf(self.recImgEnabled)
			self.recButton.set_image(image)
			self.recTip.set_tip(self.recButton, self.recTipEnabled, None)
		else:
			image.set_from_pixbuf(self.recImgDisabled)
			self.recButton.set_image(image)
			self.recTip.set_tip(self.recButton, self.recTipDisabled, None)
	
	#_____________________________________________________________________
	
	def OnInstrumentMute(self, instrument=None):
		"""
		Callback for when the instrument's muted status changes.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		self.Updating = True
		self.muteButton.set_active(self.instrument.actuallyIsMuted)
		self.Updating = False
		
		# update the mute button image and tooltip
		image = gtk.Image()
		if self.instrument.actuallyIsMuted:
			image.set_from_pixbuf(self.muteImgEnabled)
			self.muteButton.set_image(image)
			self.muteTip.set_tip(self.muteButton, self.muteTipEnabled, None)
		else:
			image.set_from_pixbuf(self.muteImgDisabled)
			self.muteButton.set_image(image)
			self.muteTip.set_tip(self.muteButton, self.muteTipDisabled, None)
	
	#______________________________________________________________________

	def OnEffectsButtonClicked(self, widget):
		"""
		Creates and shows the instrument effects dialog
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			mouse -- reserved for GTK callbacks, don't use it explicitly.
		"""
		Globals.debug("props button pressed")
		if not self.effectsDialog:
			self.effectsDialog = InstrumentEffectsDialog.InstrumentEffectsDialog(
					self.instrument,
					self.OnEffectsDialogDestroyed,
					self.mainview.icon)
		else:
			self.effectsDialog.BringWindowToFront()

	#______________________________________________________________________
	
	def OnEffectsDialogDestroyed(self, window):
		"""
		Called when the InstrumentEffectsDialog is destroyed.
		
		Parameters:
			window -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.effectsDialog = None
		
	#______________________________________________________________________
#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	LevelsList.py
#	
#	This module contains the class which is responable for storing a list of
#	all of the audio levels of each channel of audio along with the time of that
#	level within the audio track. This data is used to draw the a graphical
#	representation of the audio waveform for the user.
#
#-------------------------------------------------------------------------------

import Utils

from array import array
import itertools
import sys
import bisect
import copy

class LevelsList:
	MAGIC_NUMBER = 0x00011011	# an integer with 4 unique bytes used to check endianness
	VERSION = 1
	ARRAY_TYPE = 'l'
	
	def __init__(self):
		self.channels = []
		self.times = array(self.ARRAY_TYPE)
	
	#_____________________________________________________________________
	
	def CreateChannels(self,  num_channels):
		self.channels = []
		for i in xrange(num_channels):
			self.channels.append(array(self.ARRAY_TYPE))
	
	#_____________________________________________________________________
	
	def copy(self):
		levelslist = LevelsList()
		levelslist.times = copy.copy(self.times)
		levelslist.channels = []
		for chan in self.channels:
			levelslist.channels.append(copy.copy(chan))
	
		return levelslist

	#_____________________________________________________________________
	
	def append(self, endtime, levels):
		"""
		Append a set of waveforms to the current list,
		and associates them with the given end time.
		"""
		if not self.channels:
			self.CreateChannels(len(levels))
		
		assert len(self.channels) == len(levels)
		# make sure the endtime is greater than the previous endtime
		# if this is the first endtime, make sure its bigger than 0
		assert endtime > (self.times[-1] if self.times else 0)
		
		self.times.append(endtime)
		
		for level,  chan in itertools.izip(levels,  self.channels):
			chan.append(level)
			assert len(self.times) == len(chan)
	
	#_____________________________________________________________________
	
	def append_time_delta(self, time_delta, levels):
		if self.times:
			last_time = self.times[-1]
		else:
			last_time = 0
			
		self.append(last_time + time_delta, levels)
	
	#_____________________________________________________________________
	
	def extend(self, basetime, levelslist):
		old_length = len(self.times)
		self.times.extend(levelslist.times)
		
		# shift the new endtimes to match the length of the original audio clip
		for idx in xrange(old_length, len(self.times)):
			time = self.times[idx]
			self.times[idx] = time + basetime
		
		assert len(self.channels) == len(levelslist.channels)
		for chan, lchan in itertools.izip(self.channels, levelslist.channels):
			chan.extend(lchan)
	
	
	#_____________________________________________________________________
	
	def fromfile(self,  path):
		try:
			self.__fromfile(path)
		except (EOFError, IOError):
			# on error delete all partially loaded data
			self.channels = []
			self.times = array(self.ARRAY_TYPE)
			raise CorruptFileError()
	
	#_____________________________________________________________________
	
	def __fromfile(self,  path):
		f = open(path,  "rb")
		info = array(self.ARRAY_TYPE)
		info.fromfile(f,  4)
		
		magic,  version,  length,  num_channels = info
		
		byteswap = False
		if info[0] != self.MAGIC_NUMBER:
			info.byteswap()
			if info[0] != self.MAGIC_NUMBER:
				raise CorruptFileError("unknown endianness in levels file")
			else:
				byteswap = True
		
		#currently there is only one version
		assert version == self.VERSION
		
		self.times = array(self.ARRAY_TYPE)
		self.times.fromfile(f,  length)
		if byteswap:
			self.times.byteswap()
		
		self.CreateChannels(num_channels)
		for chan in self.channels:
			chan.fromfile(f,  length)
			if byteswap:
				self.times.byteswap()
			assert len(self.times) == len(chan)
	
	#_____________________________________________________________________
	
	def tofile(self,  path):
		f = open(path,  "wb")
		info = array(self.ARRAY_TYPE)
		info.append(self.MAGIC_NUMBER)
		info.append(self.VERSION)
		info.append(len(self.times))
		info.append(len(self.channels))
		
		info.tofile(f)
		self.times.tofile(f)
		for chan in self.channels:
			chan.tofile(f)
	
	#_____________________________________________________________________
	
	def find_endtime_index(self, time):
		return bisect.bisect_left(self.times, time)
	
	#_____________________________________________________________________
	
	def slice_by_endtime(self, starttime, stoptime=None):
		if stoptime is None:
			stop_idx = len(self.times)
		else:
			assert starttime < stoptime
			stop_idx = self.find_endtime_index(stoptime)
			
		start_idx = self.find_endtime_index(starttime)
		levelslist = LevelsList()
		levelslist.times = self.times[start_idx:stop_idx]
		
		# adjust the endtimes so they are relative to the new start time.
		for idx, time in enumerate(levelslist.times):
			levelslist.times[idx] = time - starttime
		
		levelslist.channels = []
		for chan in self.channels:
			levelslist.channels.append(chan[start_idx:stop_idx])
		
		return levelslist
	
	#_____________________________________________________________________
	
	def __iter__(self):
		# FIXME: hard coded single channel
		if not self.channels:
			self.CreateChannels(1)
		return itertools.izip(self.times,  self.channels[0])
	
	#_____________________________________________________________________
	
	def __getitem__(self, index):
		# FIXME: hard coded single channel
		return (self.times[index], self.channels[0][index])
	
	#_____________________________________________________________________
	
	def __len__(self):
		return len(self.times)
	#_____________________________________________________________________

#=========================================================================

def add(list_one, list_two):
	levelslist = list_one.copy()
	levelslist.extend(list_two)
	return levelslist

#=========================================================================
	
class CorruptFileError(EnvironmentError):
	pass

#=========================================================================





import gtk
import gobject

class MessageArea(gtk.HBox):

	__gsignals__ = {
		"response" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_INT,) ),
		"close" 	: ( gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_ACTION, gobject.TYPE_NONE, () )
	}
	
	def __init__(self):
		gtk.HBox.__init__(self)
		
		self.contents = None
		self.changing_style = False
		
		self.main_hbox = gtk.HBox(False, 16) # FIXME: use style properties 
		self.main_hbox.show()
		self.main_hbox.set_border_width(8) # FIXME: use style properties
	
		self.action_area = gtk.VBox(True, 3) # FIXME: use style properties */
		self.action_area.show()
		
		self.main_hbox.pack_end(self.action_area, False, True)
		self.pack_start(self.main_hbox, True, True)
	
		self.set_app_paintable(True)
	
		self.connect("expose-event", self.paint_message_area)
		self.connect("size-allocate", self.on_size_allocate)
	
		# Note that we connect to style-set on one of the internal
		# widgets, not on the message area itself, since gtk does
		# not deliver any further style-set signals for a widget on
		# which the style has been forced with gtk_widget_set_style() 
		self.main_hbox.connect("style-set", self.style_set)
		
	def on_size_allocate(self, widget, rectangle):
		# force a _complete_ redraw here or else in certain cases after resizing
		# some border lines are left painted on top of the main content area.
		self.queue_draw()
	
	def style_set (self, widget, prev_style):
		if self.changing_style:
			return
	
		# This is a hack needed to use the tooltip background color
		window = gtk.Window(gtk.WINDOW_POPUP)
		window.set_name ("gtk-tooltip")
		window.ensure_style()
		style = window.get_style()
	
		self.changing_style = True
		self.set_style(style)
		self.changing_style = False
	
		window.destroy()
	
		self.queue_draw()
		
	def paint_message_area(self, widget, event):
		a = widget.get_allocation()
		x = a.x + 1
		y = a.y + 1
		width = a.width - 2
		height = a.height - 2
		widget.style.paint_flat_box(widget.window, gtk.STATE_NORMAL, gtk.SHADOW_OUT, None,
		                            widget, "tooltip", x, y, width, height)
		return False
	
	def action_widget_activated (self, widget):
		resp = self.get_response_data(widget)
		if resp is None:
			resp = gtk.RESPONSE_NONE
		self.response(resp)

	def get_response_data(self, widget):
		return widget.get_data("gedit-message-area-response-data")
		
	def set_response_data(self, widget, new_id):
		widget.set_data("gedit-message-area-response-data", new_id)
		
	def add_action_widget (self, child, response_id):
		self.set_response_data (child, response_id);
	
		try:
			signal = child.get_activate_signal()
		except ValueError:
			signal = None
			
		if isinstance(child, gtk.Button):
			child.connect("clicked", self.action_widget_activated)
		elif signal:
			child.connect(signal, self.action_widget_activated)
		else:
			pass
			#g_warning ("Only 'activatable' widgets can be packed into the action area of a GeditMessageArea");
	
		if response_id != gtk.RESPONSE_HELP:
			self.action_area.pack_end(child, False, False)
		else:
			self.action_area.pack_start(child, False, False)
			
			
	def add_button (self, text, response_id):
		button = gtk.Button(stock=text)
		button.set_flags(gtk.CAN_DEFAULT)
		button.show()
		self.add_action_widget(button, response_id)
		return button

	def add_buttons(self, *buttons):
		for text, response_id in buttons:
			self.add_button(text, response_id)
			
	def set_response_sensitive(self, response_id, setting):
		children = self.action_area.get_children()
		
		for child in children:
			rd = self.get_response_data(child)
			if rd == response_id:
				child.set_sensitive(setting)
	
	def set_default_response (self, response_id):
		children = self.action_area.get_children()
		
		for child in children:
			rd = self.get_response_data(child)
			if rd == response_id:
				child.grab_default()
				
	def response (self, response_id):
		self.emit("response", response_id)
		
	def add_stock_button_with_text (self, text, stock_id, response_id):
		button = gtk.Button(text, use_underline=True)
		button.set_image(gtk.image_new_from_stock(stock_id,gtk.ICON_SIZE_BUTTON))
		button.set_flags(gtk.CAN_DEFAULT)
		button.show()
		self.add_action_widget(button, response_id)

	def set_contents(self, contents):
		self.contents = contents;
		self.main_hbox.pack_start(self.contents, True, True)
		
	def set_text_and_icon(self, icon_stock_id, primary_text, secondary_text):
		hbox_content = gtk.HBox(False, 8)
		hbox_content.show()
	
		image = gtk.image_new_from_stock(icon_stock_id, gtk.ICON_SIZE_DIALOG)
		image.show()
		hbox_content.pack_start(image, False, False)
		image.set_alignment(0.5, 0.5)
		
		vbox = gtk.VBox(False, 6)
		vbox.show()
		hbox_content.pack_start(vbox, True, True)
		
		primary_markup = "<b>%s</b>" % primary_text
		primary_label = gtk.Label(primary_markup)
		primary_label.set_use_markup(True)
		primary_label.set_line_wrap(True)
		primary_label.set_alignment(0, 0.5)
		primary_label.set_flags(gtk.CAN_FOCUS)
		primary_label.set_selectable(True)
		primary_label.show()
		
		vbox.pack_start(primary_label, True, True)

		if secondary_text:
	  		secondary_markup = "<small>%s</small>" % secondary_text
			secondary_label = gtk.Label(secondary_markup)
			secondary_label.set_flags(gtk.CAN_FOCUS)
			secondary_label.set_use_markup(True)
			secondary_label.set_line_wrap(True)
			secondary_label.set_selectable(True)
			secondary_label.set_alignment(0, 0.5)
			secondary_label.show()

			vbox.pack_start(secondary_label, True, True)
	
		self.set_contents(hbox_content);

					
gtk.binding_entry_add_signal (MessageArea, gtk.gdk.keyval_from_name("Escape"), 0, "close");
	
		